/* Program demonstrira pokretanje vise programa sekvencijalno, sa
   istim vremenskim ogranicenjem. Ako se u okviru vremenskog
   ogranicenja ne zavrsi program, onda se on prekida i pokrece naredni
   program. */
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/times.h>
#include <sys/wait.h>
#include <unistd.h>
#include <math.h>

#define EPSILON 0.001
#define BUFF_SIZE 128

/* Ove globalne promenljive koriste se samo u okviru roditeljskog procesa. */
pid_t child_pid;     // process id dete procesa
int timeout_reached; // oznacava da li je dostignut timeout ili ne

void error_fatal (const char *message){
  perror (message);
  exit (EXIT_FAILURE);
}

/* Funkcija racuna n-ti clan Fibonacijevog niza. Koristi se tip long
   double kako bi mogli da se racunaju veliki clanovi ovog niza. Kao
   argument (broj n) je mogao da se koristi i tip int, ali je ovde
   koriscen long double radi ilustracije koriscenja ovog tipa. */
long double fib (long double n){
  /* Realni brojevi se ne mogu porediti pomocu jednakosti. */
  if (fabsl (n) < EPSILON || fabsl (n-1) < EPSILON)
    return 1.0;
  else {
    long double fib_1 = fib(n-1);
    long double fib_2 = fib(n-2);
    return fib_1 + fib_2;
  }
}

/* Ako je prosao odgovarajuci broj sekundi, treba ubiti dete proces i
   postavljati promenljivu koja oznacava da je dostignuto vremensko
   ogranicenje. */
static void sig_alrm(int signo){
  timeout_reached = 1;
  if (kill (child_pid, SIGKILL) == -1) {
    fprintf(stderr, "Greska pri pozivu funkcije kill za dete sa id-om: %d\n", child_pid);
    error_fatal ("kill error");
  }
}

/* Izracunava se i-ti clan Fibonacijevog niza ukoliko je to moguce
   uraditi u okviru vremenskog ogranicenja timeout. */
void calculate_fibonaci (long double i, int timeout){
  pid_t pid;
  struct tms tms_start, tms_end;
  clock_t start, end;
  long ticks_per_second;
  int filedes[2];

  /* Pravljenje veze za slanje poruka izmedju roditelja i deteta. */
  if (pipe (filedes) < 0)
    error_fatal ("pipe error");
  
  /* Odredjuje se broj tikova u sekundi.*/
  if ( (ticks_per_second = sysconf (_SC_CLK_TCK)) <= 0)
    error_fatal ("sysconf error");

  /* Postavlja se funkcija sig_alrm da bude pozvana kada dodje do
     signala SIGALRM. Bitno je da se ovo uradi pre nego sto se pozove
     funkcija alarm.*/
  if (signal(SIGALRM, sig_alrm) == SIG_ERR)
    error_fatal ("signal error");

  /* Pokrece se hronometar (timer) na odgovarajuci broj sekundi.  
     Vrlo bitno: roditelj mora primiti ovaj signal da bi ubio dete proces.*/
  alarm (timeout);

  /* Pre izvrsavanja komande, odredjuje se broj proteklih 
     tikova od nekog (ne znamo kog) trenutka u proslosti. */
  if ( (start = times (&tms_start)) < 0)
    error_fatal ("first times call");

  /* Kreiranje novog procesa. */
  if ( (pid = fork()) < 0)
    error_fatal ("fork error");
  else if (pid == 0){
    char buffer[BUFF_SIZE];

    /* Bitno: zatvaranje deskriptora koje necemo koristiti. */
    close (filedes[0]);

    /* Izracunava se i-ti clan Fibonacijevog niza i smesta u buffer. */
    long double fib_i = fib (i);
    sprintf (buffer, "%Lf", fib_i);
    
    /* Salje se poruka roditeljskom procesu. */
    if ( (write (filedes[1], buffer, strlen (buffer) + 1)) != strlen (buffer) + 1)
      error_fatal ("write error");

    close (filedes[1]);

    /* Dete zavrsava sa radom. */
    exit (EXIT_SUCCESS);
  }
  else {
    /* Cuva se process id deteta u globalnoj promenljivoj.*/
    child_pid = pid;

    /* Bitno: zatvaranje deskriptora koje necemo koristiti.
       Ovo je najosetljivije mesto i obavezno je zatvaranje ovog deksriptora.
       U slucaju kada se ne zatvori donji read moze da blokira jer postoji
       proces koji moze da salje poruku na filedes[1]. */
    close (filedes[1]);

    /* Cekanje na dete proces. */
    if (waitpid(child_pid, NULL, 0) < 0)
      error_fatal ("waitpid error");

    /* Posle izvrsavanja komande, odredjuje se broj proteklih 
       tikova od nekog (ne znamo kog) trenutka u proslosti. */
    if ( (end = times (&tms_end)) < 0)
      error_fatal ("second times call");

    /* Stampanje rezultata ako je postignut ili pak poruke da nije. */
    if (timeout_reached == 0){

      /* Citanje poruke od deteta procesa. */
      char buffer[BUFF_SIZE];
      if (read (filedes[0], buffer, sizeof (buffer)) < 0)
	error_fatal ("read error");
      
      /* Naredna konverzija je potencijalno problematicna, jer se
	 koriscenjem atof mogu izgubiti neki podaci - ova funkcija
	 vraca double a ne long double. */
      long double fib_read = atof (buffer);
      printf ("fib (%Lf) = %Lf  (%g seconds)\n", i, fib_read, (double)(end - start) / ticks_per_second);
    }
    else
      printf ("Not found in %g seconds\n", (double)(end - start) / ticks_per_second);

    /* Resetovanje varijable timeout_reached. */
    timeout_reached = 0;
    close (filedes[0]);
  }
}
int main (int argc, char **argv){
  int i;
  int timeout;
  int max_n = atoi (argv[1]);

  /* Provera argumenata komandne linije. */
  if (argc != 3){
    fprintf (stderr, "./fibonaci_timer maximal_index_of_fibonaci_array timeout\n");
    exit (EXIT_FAILURE);
  }

  timeout = atoi (argv[2]);

  for (i=0; i<max_n; i++)
    calculate_fibonaci (i+0.0, timeout);

  return 0;
}

